#!/usr/local/bin/python
# -*- coding: utf-8 -*-

__version__ = '1.0.11'

import logging

from slashdb.app import create_slashdb_app
import sqlalchemy
from sqlalchemy import util
from sqlalchemy.dialects import oracle
from sqlalchemy.dialects.sqlite import base
from sqlalchemy.engine.default import DefaultExecutionContext
from sqlalchemy.ext import automap, declarative
from sqlalchemy.sql import expression
from sqlalchemy.orm import interfaces, relationship, backref
from sqlalchemy.sql.schema import ColumnDefault, ForeignKeyConstraint


log = logging.getLogger('slashdb')


def oracle_patches():
    u"""SQLite Oracle monkey patches."""

    def patched_is_many_to_many(automap_base, table):
        return None, None, None
    # patched function because in SlashDB we don't want to skip junction tables.
    automap._is_many_to_many = patched_is_many_to_many

    base.ischema_names['NVARCHAR'] = sqlalchemy.types.VARCHAR
    oracle.base.ischema_names['NCHAR'] = oracle.NVARCHAR2
    oracle.NCHAR = oracle.NVARCHAR2


def patch_mysql_compound_pk():
    u"""Patches for issue #360.

    Code taken from https://bitbucket.org/zzzeek/sqlalchemy/issues/3921

    """

    @util.memoized_property
    @util.dependencies("sqlalchemy.sql.sqltypes")
    def _arg_is_typed(self, sqltypes):
        if self.is_clause_element:
            # need to resolve the lazy impport
            # otherwise we'll get:
            #   ImportError: importlater.resolve_all() hasn't been called (this is sqlalchemy.sql sqltypes)
            sqltypes._resolve()
            return not isinstance(self.arg.type, sqltypes.NullType)
        else:
            return False

    ColumnDefault._arg_is_typed = _arg_is_typed

    def _exec_default(self, column, default, type_):
        if default.is_sequence:
            return self.fire_sequence(default, type_)
        elif default.is_callable:
            self.current_column = column
            return default.arg(self)
        elif default.is_clause_element:
            # TODO: expensive branching here should be
            # pulled into _exec_scalar()
            conn = self.connection
            if not default._arg_is_typed:
                default_arg = expression.type_coerce(default.arg, type_)
            else:
                default_arg = default.arg
            c = expression.select([default_arg]).compile(bind=conn)
            return conn._execute_compiled(c, (), {}).scalar()
        else:
            return default.arg

    DefaultExecutionContext._exec_default = _exec_default


def patch_db2_automap():
    u"""Patches for issue #444.

    Prevents problems when reflecting DB2 with automap.

    """

    def clean_fks(fks):
        clean_fks = {fk.name: fk for fk in fks if fk.parent is not None}
        clean_fk = None
        for fk in fks:
            if fk.name in clean_fks and fk.parent is None:
                clean_fk = clean_fks[fk.name]
                fk.parent = clean_fk.parent
                fk.column = clean_fk.column
        return fks

    def _relationships_for_fks(automap_base, map_config, table_to_map_config,
                               collection_class,
                               name_for_scalar_relationship,
                               name_for_collection_relationship,
                               generate_relationship):
        local_table = map_config.local_table
        local_cls = map_config.cls

        if local_table is None:
            return
        for constraint in local_table.constraints:
            if isinstance(constraint, ForeignKeyConstraint):
                fks = clean_fks(constraint.elements)
                referred_table = fks[0].column.table
                referred_cfg = table_to_map_config.get(referred_table, None)
                if referred_cfg is None:
                    continue
                referred_cls = referred_cfg.cls

                if local_cls is not referred_cls and issubclass(
                        local_cls, referred_cls):
                    continue

                relationship_name = name_for_scalar_relationship(
                    automap_base,
                    local_cls,
                    referred_cls, constraint
                )
                backref_name = name_for_collection_relationship(
                    automap_base,
                    referred_cls,
                    local_cls,
                    constraint
                )

                o2m_kws = {}
                nullable = False not in set(fk.parent.nullable for fk in fks)
                if not nullable:
                    o2m_kws['cascade'] = "all, delete-orphan"

                    if constraint.ondelete and \
                            constraint.ondelete.lower() == "cascade":
                        o2m_kws['passive_deletes'] = True
                else:
                    if constraint.ondelete and \
                            constraint.ondelete.lower() == "set null":
                        o2m_kws['passive_deletes'] = True

                create_backref = backref_name not in referred_cfg.properties

                if relationship_name not in map_config.properties:
                    if create_backref:
                        backref_obj = generate_relationship(
                            automap_base,
                            interfaces.ONETOMANY, backref,
                            backref_name, referred_cls, local_cls,
                            collection_class=collection_class,
                            **o2m_kws)
                    else:
                        backref_obj = None
                    foreign_keys = [fk.parent for fk in fks]
                    remote_side = [fk.column for fk in fks]
                    rel = generate_relationship(
                        automap_base, interfaces.MANYTOONE, relationship,
                        relationship_name, local_cls, referred_cls,
                        foreign_keys=foreign_keys, backref=backref_obj,
                        remote_side=remote_side
                    )
                    if rel is not None:
                        map_config.properties[relationship_name] = rel
                        if not create_backref:
                            referred_cfg.properties[
                                backref_name].back_populates = relationship_name
                elif create_backref:
                    foreign_keys = [fk.parent for fk in fks]
                    rel = generate_relationship(
                        automap_base, interfaces.ONETOMANY, relationship,
                        backref_name, referred_cls, local_cls,
                        foreign_keys=foreign_keys, back_populates=relationship_name,
                        collection_class=collection_class, **o2m_kws
                    )
                    if rel is not None:
                        referred_cfg.properties[backref_name] = rel
                        map_config.properties[relationship_name].back_populates = backref_name

    automap._relationships_for_fks = _relationships_for_fks

    class AutomapBase(object):
        """Base class for an "automap" schema.

        The :class:`.AutomapBase` class can be compared to the "declarative base"
        class that is produced by the :func:`.declarative.declarative_base`
        function.  In practice, the :class:`.AutomapBase` class is always used
        as a mixin along with an actual declarative base.

        A new subclassable :class:`.AutomapBase` is typically instantated
        using the :func:`.automap_base` function.

        .. seealso::

            :ref:`automap_toplevel`

        """
        __abstract__ = True

        classes = None
        """An instance of :class:`.util.Properties` containing classes.

        This object behaves much like the ``.c`` collection on a table.  Classes
        are present under the name they were given, e.g.::

            Base = automap_base()
            Base.prepare(engine=some_engine, reflect=True)

            User, Address = Base.classes.User, Base.classes.Address

        """

        @classmethod
        def prepare(
                cls,
                engine=None,
                reflect=False,
                classname_for_table=automap.classname_for_table,
                collection_class=list,
                name_for_scalar_relationship=automap.name_for_scalar_relationship,
                name_for_collection_relationship=automap.name_for_collection_relationship,
                generate_relationship=automap.generate_relationship):
            """Extract mapped classes and relationships from the :class:`.MetaData` and
            perform mappings.

            :param engine: an :class:`.Engine` or :class:`.Connection` with which
            to perform schema reflection, if specified.
            If the :paramref:`.AutomapBase.prepare.reflect` argument is False,
            this object is not used.

            :param reflect: if True, the :meth:`.MetaData.reflect` method is called
            on the :class:`.MetaData` associated with this :class:`.AutomapBase`.
            The :class:`.Engine` passed via
            :paramref:`.AutomapBase.prepare.engine` will be used to perform the
            reflection if present; else, the :class:`.MetaData` should already be
            bound to some engine else the operation will fail.

            :param classname_for_table: callable function which will be used to
            produce new class names, given a table name.  Defaults to
            :func:`.classname_for_table`.

            :param name_for_scalar_relationship: callable function which will be
            used to produce relationship names for scalar relationships.  Defaults
            to :func:`.name_for_scalar_relationship`.

            :param name_for_collection_relationship: callable function which will
            be used to produce relationship names for collection-oriented
            relationships.  Defaults to :func:`.name_for_collection_relationship`.

            :param generate_relationship: callable function which will be used to
            actually generate :func:`.relationship` and :func:`.backref`
            constructs.  Defaults to :func:`.generate_relationship`.

            :param collection_class: the Python collection class that will be used
            when a new :func:`.relationship` object is created that represents a
            collection.  Defaults to ``list``.

            """
            if reflect:
                cls.metadata.reflect(
                    engine,
                    extend_existing=True,
                    autoload_replace=False
                )

            table_to_map_config = dict(
                (m.local_table, m)
                for m in automap._DeferredMapperConfig.
                classes_for_base(cls, sort=False)
            )

            many_to_many = []

            for table in cls.metadata.tables.values():
                lcl_m2m, rem_m2m, m2m_const = automap._is_many_to_many(cls, table)
                if lcl_m2m is not None:
                    many_to_many.append((lcl_m2m, rem_m2m, m2m_const, table))
                elif not table.primary_key:
                    continue
                elif table not in table_to_map_config:
                    mapped_cls = type(
                        classname_for_table(cls, table.name, table),
                        (cls, ),
                        {"__table__": table}
                    )
                    map_config = declarative.base._DeferredMapperConfig.config_for_cls(mapped_cls)
                    cls.classes[map_config.cls.__name__] = mapped_cls
                    table_to_map_config[table] = map_config

            for map_config in table_to_map_config.values():
                _relationships_for_fks(cls,
                                       map_config,
                                       table_to_map_config,
                                       collection_class,
                                       name_for_scalar_relationship,
                                       name_for_collection_relationship,
                                       generate_relationship)

            for lcl_m2m, rem_m2m, m2m_const, table in many_to_many:
                automap._m2m_relationship(cls, lcl_m2m, rem_m2m, m2m_const, table,
                                          table_to_map_config,
                                          collection_class,
                                          name_for_scalar_relationship,
                                          name_for_collection_relationship,
                                          generate_relationship)

            for map_config in declarative.base._DeferredMapperConfig.classes_for_base(cls):
                map_config.map()

        _sa_decl_prepare = True
        """Indicate that the mapping of classes should be deferred.

        The presence of this attribute name indicates to declarative
        that the call to mapper() should not occur immediately; instead,
        information about the table and attributes to be mapped are gathered
        into an internal structure called _DeferredMapperConfig.  These
        objects can be collected later using classes_for_base(), additional
        mapping decisions can be made, and then the map() method will actually
        apply the mapping.

        The only real reason this deferral of the whole
        thing is needed is to support primary key columns that aren't reflected
        yet when the class is declared; everything else can theoretically be
        added to the mapper later.  However, the _DeferredMapperConfig is a
        nice interface in any case which exists at that not usually exposed point
        at which declarative has the class and the Table but hasn't called
        mapper() yet.

        """

    automap.AutomapBase = AutomapBase


patches = (oracle_patches, patch_mysql_compound_pk)

for patch in patches:
    patch()

log.debug('Applied patches {}'.format(', '.join(patch.__name__ for patch in patches)))


def main(global_config, **settings):
    """This function returns a Pyramid WSGI application.

    :param global_config:
    :param settings: dictionary of settings passed from INI file

    """

    return create_slashdb_app(global_config, settings)
