class QuerystudioController
  @$inject: ['$window', '$location', '$timeout', '$rootScope', '$scope', 'slashDB', 'QueryExecService', 'selectizeService']
  constructor: (@$window, @$location, @$timeout, @$rootScope, @$scope, @slashdb, @QES, @selectizeService) ->
    @tabs =
      side:
        config: true
        schema: false
      top:
        sql: true
        diagram: false
      bottom:
        execute: true
        results: false
    @required_fields = {}
    @executeOnly = null
    @codemirrorInstance = null
    @lastDbSchemaId = null
    @execUrls =
      html: ''
      json: ''
      csv: ''
      xml: ''
      xsd: ''

  $onInit: ->
    @slashdb.setupSettings()
    .then =>
      @initCodemirror()
      @initConfigForm()
      @initExecuteForm()
      @initResults()
      @initSchema()
      @updateConfigForm()
      .then =>
        @updateCodemirror()
        executeNow = @restoreExecuteForm()
        if executeNow
          @updateExecuteForm(@configData)
          res = @executeQuery()
          if not res
            @switchTab('bottom', 'execute')

    return null

  initSchema: ->
    @$scope.$watch =>
      return @configData.database
    , (newValue, oldValue) =>
      if newValue and newValue != oldValue
        @updateHints(newValue)

  updateHints: (dbId) ->
    url = "/db/#{dbId}.json?tables"
    @slashdb.get url
    .then (response) =>
      schemaTables = response.data
      schemaHints = new Set([]);

      for table in schemaTables
        schemaHints.add table.name
        for column in table.columns
          schemaHints.add column.name

      @schemaHints = Array.from(schemaHints)

  initConfigForm: ->
    @slashdb.get('/db.json', { params: guidata: '' }, false, false).then (response) =>
      @$scope.databases = response.data
      return
    # stores database definition model, some initial values
    @configData = editable: true

    @updateSuggestedMethods()
    @$scope.newConfig = !Boolean(@queryId)
    @$scope.configId = @queryId
    @$scope.okVisible = false
    @$scope.errorMessage = ''
    @$scope.databases = []


  initExecuteForm: ->
    @httpMethods = ['GET', 'POST', 'PUT', 'DELETE']
    @extensions = ['html', 'json', 'csv', 'xml', 'xsd']

    @defaultLimit = parseInt(@slashdb.settings.html_limit) or 29

    @ready = false
    @limit = @defaultLimit
    @offset = 0
    @count = false
    @format = 'json'
    @execUrl = ''
    @execPath = ''
    @queryExecuted = false
    @selectedHTTPMethod = null
    @queryParameters = {}

    @$scope.methodAvailable = (m) =>
      http_methods = @configData.http_methods
      if http_methods
        return http_methods[m]
    @$scope.hasParameters = =>
      return Object.keys(@queryParameters).length > 0

  initResults: ->
    @queryOutput = null

    @$rootScope.$on 'queryExecuted', (event, args) =>
      @queryOutput = @QES.queryOutput
      httpResponse = args.httpResponse
      @resultsResponseCode = httpResponse.status
      @resultsResponseMessage = "#{httpResponse.status} #{httpResponse.statusText}"
      @resultsResponseWarning = httpResponse.headers 'Warning'
      @resultsCount = httpResponse.headers(
        "#{@slashdb.settings.branding_name}-all-record-count"
      )
      @resultsCountFailed = @resultsResponseWarning == 'Failed to get count.'

      noData = (httpResponse.data? and httpResponse.data.length == 0) or not httpResponse.data?
      if httpResponse.status < 400 and noData
        @resultsResponseNoMoreData = true
      else
        @resultsResponseNoMoreData = false


      @switchTab 'bottom', 'results'

  initCodemirror: ->
    handleLoadedEditor = (_editor) =>
      @codemirrorInstance = _editor
      _editor.on 'change', =>
        @updateSuggestedMethods()

    produceHints = (mc, opts) =>
      hints = CodeMirror.hint.sql(mc, opts)
      dbschemaHints = CodeMirror.hint.dbschema(mc, opts, @schemaHints)

      if dbschemaHints
        for hint in dbschemaHints.list
          hints.list.push hint

      return hints

    # Has to change in value to apply changes in CodeMirror settings
    @$scope.refreshCodemirror = true
    @$scope.editorOptions =
      onLoad: handleLoadedEditor
      lineNumbers: true
      readOnly: false
      mode: 'text/x-sql'
      hintOptions:
        hint: produceHints
      extraKeys:
        "Ctrl-Space": "autocomplete"

  updateCodemirror: ->
    currDbType = @$scope.databases?[@configData.database]?.db_type
    cmMode = switch currDbType
      when 'sqlite' then "text/x-sqlite"
      when 'mssql' then 'text/x-mssql'
      when 'oracle' then 'text/x-plsql'
      when 'mysql' then 'text/x-mysql'
      when 'postgresql' then 'text/x-gpsql'
      when 'db2' then 'text/x-sql'
      else 'text/x-sql'

    @$scope.editorOptions.mode = cmMode
    @$scope.editorOptions.readOnly = if !@configData?.editable then 'nocursor' else false
    @$scope.refreshCodemirror = !@$scope.refreshCodemirror

  switchTab: (section, tabName) ->
    for tab in (Object.keys @tabs[section])
      @tabs[section][tab] = tab == tabName

  updateConfigForm: ->
    if @$scope.newConfig
      @executeOnly = false
      @configData.executable = true
      @updateCodemirror()
      @selectizeService.updateUserOptions null
    else
      @$scope.errorMessage = ''
      promise = @slashdb.get '/querydef/' + @$scope.configId + '.json', { params: guidata: '' }, false, false
      .then (response) =>
        @configData = angular.copy(response.data)
        @executeOnly = @configData.executable and !(@configData.editable or @configData.viewable)
        if @configData.editable or @configData.viewable
          @selectizeService.updateUserOptions @configData

        @updateSuggestedMethods()
        @updateCodemirror()
        @updateExecuteForm(@configData)
        return
      .catch (response) =>
          @$scope.errorMessage = response.headers('Warning')

      return promise

  clean: (andDisable) ->
    @configData = editable: !andDisable
    return

  delete: ->
    @$scope.errorMessage = ''
    @slashdb.deleteQueryDef(@$scope.configId)
    .then (response) =>
      @slashdb.setupSettings()
      @$window.location.href = '/querydef'
      return
    .catch (response) =>
      @$scope.errorMessage = response.headers('Warning')
      return
    return

  saveExecuteForm: (execute)->
    formParams =
      format: @format
      limit: @limit
      offset: @offset
      count: @count
      selectedHTTPMethod: @selectedHTTPMethod
      queryParameters: @queryParameters
      queryId: @configData.query_id
      executeNow: execute

    try
      localStorage.setItem('qsExecuteQuery', JSON.stringify(formParams))
    catch e
      if e == QUOTA_EXCEEDED_ERR
        alert('local storage limit exeeded')

  restoreExecuteForm: ->
    formParams = JSON.parse(localStorage.getItem('qsExecuteQuery'))
    if not formParams?
      return false
    localStorage.removeItem('qsExecuteQuery')

    if formParams.queryId != @queryId
      return false

    @limit = formParams.limit
    @offset = formParams.offset
    @count = formParams.count
    @format = formParams.format
    @selectedHTTPMethod = formParams.selectedHTTPMethod
    @queryParameters = formParams.queryParameters

    return formParams.executeNow

  save: (execute = false) ->
    if not @configData.read then @configData.read = []
    if not @configData.write then @configData.write = []
    if not @configData.execute then @configData.execute = []

    @$scope.errorMessage = ''
    if @$scope.newConfig
      promise = @slashdb.createQueryDef(@configData)
      promise.then (response) =>
        @saveExecuteForm(execute)
        @goToNewEndpoint()
        return
      .catch (response) =>
        @$scope.errorMessage = response.headers('Warning')
        if response.data.required_fields?
          for name in response.data.required_fields
            @required_fields[name] = true
          return
    else
      dirtyUrl = @configForm.query_id.$dirty
      promise = @slashdb.updateQueryDef(@$scope.configId, @configData)
      promise.then (response) =>
        @blinkOK()
        @$scope.configId = @configData.query_id
        @updateConfigForm()
        if dirtyUrl
          @saveExecuteForm(execute)
          @goToNewEndpoint()
        return
      .catch (response) =>
        @$scope.errorMessage = response.headers('Warning')
        if response.data.required_fields?
          for name in response.data.required_fields
            @required_fields[name] = true
        return

    return promise

  goToNewEndpoint: ->
    @$window.location.assign("/querystudio/#{@configData.query_id}")

  saveAndExecute: ->
    @save(execute=true)
    .then (sdbResponse) =>
      if not @$scope.newConfig and sdbResponse.status >= 200 and sdbResponse.status < 400
        @updateConfigForm()
        .then =>
          res = @executeQuery()
          if not res
            @switchTab('bottom', 'execute')

  previousResultPage: ->
    @offset -= @limit
    if @offset < 0 then @offset = 0
    @executeQuery()
    return

  nextResultPage: ->
    @offset += @limit
    @executeQuery()
    return

  updateExecUrl: ->
    if @configData
      @validate()
      if @ready
        for key of @execUrls
          urlPath = @QES.getExecPath @configData.query_id, @queryParameters, @limit, @offset, @count, key
          @execUrls[key] = @QES.queryBasePath + urlPath
      else
        for key of @execUrls
          @execUrls[key] = ''

      url = @QES.getExecPath @configData.query_id, @queryParameters, @limit, @offset, @count, @format
      @execUrl = @QES.queryBasePath + url
      @execPath = url
      return @execUrl
    return

  executeQuery: ->
    @validate()
    if @ready

      queryStringParams =
        limit: @limit
        offset: @offset
        columns: true
        numeric_columns: true
      if @count
        queryStringParams['count']=''

      if Object.keys(@queryParameters).length == 0
        queryParams = null
      else if @selectedHTTPMethod in ['PUT', 'POST']
        queryParams = [@queryParameters]
      else
        queryParams = @queryParameters

      @QES.executeQuery @configData.query_id, @selectedHTTPMethod, queryParams, queryStringParams
    return @ready

  validate: ->
    valid = true
    valid = valid and @limit > 0
    valid = valid and @format in @extensions
    valid = valid and @selectedHTTPMethod in @httpMethods

    if @executeForm
      for key of @queryParameters
        formField = @executeForm[key]
        if formField
          formField.$touched = true
          valid = valid and formField.$valid
    # page not loaded yet, just check that params are present
    else
      for key, value of @queryParameters
        valid = valid and value

    @ready = valid

  blinkOK: ->
    @$scope.okVisible = true
    @$timeout =>
      @$scope.okVisible = false
      return
    , 1000
    return

  methodsMapping =
    'select':
      re: /\bselect\b/i
      method: 'GET'
    'insertOrExec':
      re: /\binsert|exec\b/i
      method: 'POST'
    'update':
      re: /\bupdate\b/i
      method: 'PUT'
    'delete':
      re: /\bdelete\b/i
      method: 'DELETE'
  methodsMappingKeys = Object.keys(methodsMapping)
  methodsMappingKeysL = methodsMappingKeys.length

  updateSuggestedMethods: ->
    suggestedMethods =
      'GET': false
      'POST': false
      'PUT': false
      'DELETE': false
    if @configData == null
      return suggestedMethods
    sqlstr = @configData.sqlstr
    searchTerm = undefined
    i = 0
    while i < methodsMappingKeysL
      searchTerm = methodsMapping[methodsMappingKeys[i]]
      suggestedMethods[searchTerm.method] = searchTerm.re.exec(sqlstr) != null
      i++
    @$scope.suggestedMethods = suggestedMethods

  updateExecuteForm: (query) ->
    re = /:([A-Za-z_]\w*)/g
    oldParameters = angular.copy @queryParameters
    availableMethods = (m for m, v of query.http_methods when v is true)
    @updateSuggestedMethods()
    @queryParameters = {}

    # only when user have view access
    # params from editor
    if query.sqlstr
      match = re.exec query.sqlstr
      while match
        paramName = match[1]
        if oldParameters.hasOwnProperty paramName
          paramValue = oldParameters[paramName]
        else
          paramValue = ''
        @queryParameters[paramName] = paramValue
        match = re.exec query.sqlstr

    # when user do not have view access
    # params from API
    if query.parameters
      for paramName in query.parameters
        if oldParameters.hasOwnProperty paramName
          paramValue = oldParameters[paramName]
        else
          paramValue = ''

        @queryParameters[paramName] = paramValue

    # preselect some http method for the form. First select by suggested then by available.
    if @selectedHTTPMethod not in availableMethods
      for m in @httpMethods
        if m in availableMethods
          if @$scope.suggestedMethods[m]
            @selectedHTTPMethod = m
            break

    if @selectedHTTPMethod not in availableMethods
      for m in @httpMethods
        if m in availableMethods
          @selectedHTTPMethod = m
          break

    return

  newQuery: ->
    if @configForm.$dirty or @sqlDirty
      @confirmNewQueryVisible = true
    else
      @openNewQueryPage()

  openNewQueryPage: ->
      window.location.assign("/querystudio/")

  sqlKeyDown: (e) ->
    if @configData.executable and (e.ctrlKey or e.metaKey) and (e.keyCode == 13 or e.keyCode == 10)
      @saveAndExecute()

  sqlChange: (e) ->
    if @configData.executable
      @updateExecuteForm(@configData)
      @sqlDirty = true

  onHttpMethodChange: ->
    if @configData.executable
      @updateExecuteForm(@configData)


angular.module 'sdbgui'
.component 'querystudio', {
  templateUrl: '/static/components/sdbgui-querystudio/querystudio.html'
  controller: QuerystudioController
  bindings:
    queryId: '@'
}
