# -*- coding: utf-8 -*-

from UserDict import DictMixin
import abc
import collections

MODULE_MAKERS = {}
DATA_PROXIES = {}
DB_CONFIGS = {}


class ConnectStatus:
    """Class attribute which indicate database connection state while reflecting."""

    ON = "Connected"
    OFF = "Disconnected"
    CONNECTING = "Connecting"
    FAILED = "Failed"


class AttrType:
    """Class attribute which identifies types of object attributes.

    It's needed by renderer to distinguish between columns,
    many-to-one relation and one-to-many relation.

    Used by IModel.get_attr_type method.

    """

    COLUMN = 1
    TOONE = 2
    TOMANY = 3


class IModuleMaker(object):
    """Database model interface."""

    def __init__(self, db_id, db_config, ini_settings):
        """
        :param db_id - a string representing the data sources unique identifier
        :param db_config - a dict-like object, representing the data sources configuration
        :param ini_settings - a dictionary representing the parsed configuration *.ini file
        """
        self.db_id = db_id
        self.db_config = db_config
        self.ini_settings = ini_settings

    def create(self):
        """Returns a new python module with classes that have implemented IModel interface.

        Basically
        imp.new_module(self.db_id)
        new_class = type(str(class_name), (ModelClass,), {'_field_names': field_names})
        setattr(new_module, new_class.__name__, new_class)
        return new_module

        In above short example
        class_name - name of a new class
        ModelClass - class template that implements IModel from plugin package
        {'_field_names': field_names} - class attributes of new class

        """
        raise NotImplementedError


class IModel(DictMixin, object):

    """Basic resource/model interface.

    Reflected models are expected to implement these methods for proper rendering.
    Some methods are required by DictMixin to make this object behave like a dictionary.
    """

    def __getitem__(self, param):
        raise NotImplementedError

    def __setitem__(self, key, value):
        raise NotImplementedError

    def keys(self):
        raise NotImplementedError

    def iteritems(self):
        raise NotImplementedError

    def iterkeys(self):
        raise NotImplementedError

    def get_href(self, field_name=None, slash_escaping_fun=None):
        """Returns href without domain to resource or field or relation of the resource."""
        raise NotImplementedError

    def get_href_segments(self, field_name=None, slash_escaping_fun=None):
        """Returns href without domain to resource or field or relation of the resource as list of url segments."""
        raise NotImplementedError

    @classmethod
    def get_arranged_props(cls):
        """List of strings with names of attributes/keys in preferred order (columns and relations)."""
        raise NotImplementedError

    @classmethod
    def get_column_names(cls):
        """List of string with names of attributes/keys that refer to columns."""
        raise NotImplementedError

    @classmethod
    def get_relation_names(cls):
        """List of strings with names of attributes/keys that refer to relationships in object."""
        raise NotImplementedError

    @classmethod
    def get_related_class(cls, relation_name):
        """Returns a related class by given name of the relation (object attribute)."""
        raise NotImplementedError

    @classmethod
    def get_python_type(cls, column_name):
        """Returns a correct python type for column data."""
        raise NotImplementedError

    @classmethod
    def get_python_value(cls, column_name, text_value):
        """Converts text to proper python type value."""
        raise NotImplementedError

    @classmethod
    def get_attr_type(cls, attr_name):
        """Returns proper slashdb.models.AttrType for queried attribute/key."""
        raise NotImplementedError

    def __repr__(self):
        return '<{}>'.format(type(self).__name__)


class IDataProxy(object):
    """Data proxy interface.

    Objects of that class implement functionality of
    getting, streaming, updating, creating and deleting source data.

    Attributes:
        db_types (list): list of strings representing database types supported by this data proxy.
            This also works as an identifier in license.

        is_single_resource (bool): True if present output as single resource
            For instance when PK filtering is applied.

        resource_context: slashdb.datastructures.ResourceMeta object with details for the renderers
            ResourceMeta(
                resource_class=TheClass,        # class of IModel
                scalars=['field_a', 'firld_b']  # list of strings with names of fields for vector or array
            )

    """

    db_types = None
    is_single_resource = None
    resource_context = None

    def __init__(self, context, request):
        """
        Args:
            context: instance of slashdb.roots.DataDiscovery of slashdb.roots.Query.
                Context provides
                    context.url_segments - an interable of information caring parts of the url i.e. ('SomeDB', 'MyTable', 'MyColumn')
                    context.db_id - database unique ID
                    context.db_config - a dict-like object, representing the data sources configuration
                    context.query_id  - it's None if used DataDiscovery, way check if SQL Pass-thru
                    context.query_config  - it's None if used DataDiscovery
                    context.authenticated_user_id - ID of the logged-in user 
                    context.authenticated_user_config - a dict-like object, representing the users configuration
                    context.settings  - settings prepared on start. Consider as deprecated.
                        This should not be used as trying to remove it from application.
                    context.format  - requested output format based on extension and mime type.
                        This should be bound to available renderers formats.
                        Right now this is used as key to pick renderer or parser from
                        slashdb.formats.RENDERES or slashdb.formats.PARSERS.

            request (pyramid.request.Request): pyramid request which is source of URL params, and ini settings

        """
        self.context = context
        self.request = request

    def get(self, stream, flatten):
        """Method called by renderers to acquire data.

        Streaming data is optional allows to query data in chunks and render
        sequentially by renderer. That would reduce memory usage.

        Args:
            stream (bool): if True an iterator will be returned instead of list
            flatten (bool): if True the returned data will be flattened
        """
        raise NotImplementedError

    def create(self, data):
        """Method that will be called to save new data into the resource.

        It is called by slashdb.data_discovery_view.post_view. Together with information from context
        it should be sufficient to commit new data to database.

        Args:
            data (dict): with data for new resource

        """
        raise NotImplementedError

    def update(self, update_data):
        """Method that will be called to update data in the resource.

        It is called by slashdb.data_discovery_view.put_view. Together with information from context
        it should be sufficient to commit new data to database.

        Args:
            update_data (dict): with data for new resource

        """
        raise NotImplementedError

    def delete(self):
        """Method that will be called to delete data from the resource."""
        raise NotImplementedError

    def close(self):
        """Method to be called when we are done with using this resource.

        Useful when there's a need to do nice cleanup on the db session or file.
        Pyramid calls app_iter.close() which should call data_proxy.close().

        """

    def response_postprocessing(self, response):
        """Method for modifying freshly created response.

        Allows data proxy to make some changes in response like add a header.

        Args:
            response (pyramid.response.Response): created response object
                with defined body or app_iter, content type and charset


        Returns:
             None

        """
        pass


class CleanBase(object):
    u'''A simple class, adding field change detection.'''

    _keys = ()
    _clean = True
    # special value to distinguish between a situation where a
    # field has been explicitly set to None
    # and when getattr returns the default None
    # i.e. field was not set at all
    _not_yet_set = 'NotYetSet'

    def __setattr__(self, name, value, nys=_not_yet_set):
        old_value = getattr(self, name, nys)
        if old_value != nys and name in self._keys:
            if value != old_value:
                self._clean = False
            else:
                return
        super(CleanBase, self).__setattr__(name, value)

    def __delattr__(self, name):
        if name in self._keys:
            self.clean = False
        super(CleanBase, self).__delattr__(name)

    @property
    def is_dirty(self):
        u"""Check if fields declared in _keys were modified."""
        return not self._clean

    def set_clean(self):
        u"""Sets instance state to clean."""
        self._clean = True


class ValidationMixin(object):
    def __init__(self, *args, **kwargs):
        # validation attributes
        self.errors = []
        self.warnings = []
        self.required_fields = []
        self.conflict = False

    def add_error(self, msg):
        self.errors.append(msg)

    def add_warning(self, msg):
        self.warnings.append(msg)

    def add_required(self, *args):
        for field_name in args:
            self.required_fields.append(field_name)

    def validate(self, **kwargs):
        """Validates config data.

        :param kwargs: a dictionary containing any additional info that the validate method might want to use

        """
        self.reset_validation()
        return self.errors, self.warnings

    def reset_validation(self):
        """Clears validation result."""
        self.errors = []
        self.warnings = []
        self.required_fields = []
        self.conflict = False


class IDBConfig(CleanBase, ValidationMixin):

    """Abstract class for SlashDB configuration objects."""

    __metaclass__ = abc.ABCMeta

    _log_type = 'database'
    _keys = [
        'db_id', 'db_type', 'db_encoding', 'desc', 'autoload', 'autoconnect',
        'creator', 'read', 'write', 'execute',
        'connection', 'db_schema', 'autoload_user', 'alternate_key',
        'excluded_columns', 'foreign_keys', 'connect_status'
    ]

    def __init__(self, db_id, db_type, db_encoding, desc, autoload, autoconnect,
                 creator, read, write, execute,
                 connection, db_schema, autoload_user, alternate_key, excluded_columns, foreign_keys,
                 connect_status=None):
        """Sets up DBConfig object.

        Validation can be skipped with validate=False if you're sure that attribute values are correct.

        :param db_id: model, database id
        :param db_type: type like sqlite, mssql etc.
        :param connection: connection string without db_type
        :param db_encoding: database charset encoding
        :param desc: more detail description of the database
        :param autoload_user: database user and password
        :param autoload: if True then SlashDB will try to reflect database
        :param autoconnect: if True then SlashDB will connect database automatically on start
        :param creator: id of user that added this database config
        :param read: users allowed to read this database config
        :param write: users allowed to read and change this database config
        :param execute: users allowed to connect/disconnect this database on demand
        :param db_schema: database schema
        :param alternate_key: list of columns to use as composite primary key (to be used with autoload=True)
        :param excluded_columns: list of columns to exclude from table (to be used with autoload=True)
        :param connect_status: state offline/connecting/connected/failed
        :param foreign_keys: a dict of user defined relations

        """
        
        super(IDBConfig, self).__init__()

        self.db_id = db_id
        self.db_type = db_type
        self.connection = connection
        self.db_encoding = db_encoding
        self.desc = desc
        self.autoload_user = autoload_user
        self.autoload = autoload
        self.autoconnect = autoconnect
        self.creator = creator
        self.read = read
        self.write = write
        self.execute = execute
        self.db_schema = db_schema
        self.alternate_key = alternate_key
        self.excluded_columns = excluded_columns
        self.foreign_keys = foreign_keys if foreign_keys else {}
        self.connect_status = connect_status or ConnectStatus.OFF

        self.errors = []
        self.warnings = []

    def __unicode__(self):
        return u"<{}>".format(type(self).__name__)

    def __repr__(self):
        return "<{}>".format(type(self).__name__)

    @property
    def simple_dict(self):
        return {key: getattr(self, key) for key in self._keys}

    def get_storage_dict(self, skip=None):
        skip = skip or []
        storage_dict = self.simple_dict
        for key in skip:
            storage_dict.pop(key, None)
        return storage_dict

    @property
    def is_connected(self):
        return self.connect_status == ConnectStatus.ON

    def _extract_users(self, users_field, field_name):
        """Converts context of users_fields into list of strings.

        :param users_field: contains names of users
        :type users_field: string, set or list
        :param field_name: field of database config that represent user group or permissions like (view, edit, execute).
        :type field_name: str

        :returns: list of user names
        :rtype: list of str

        """
        u_list, is_invalid_type = parse_as_list(users_field)
        if is_invalid_type:
            self.add_error('Field {} must contain comma separated user names.'.format(field_name))

        return u_list

    def _validate_db_id(self, *args, **kwargs):
        # fix circular import
        from slashdb.common import regex_good_name

        if not isinstance(self.db_id, basestring) or not self.db_id.strip():
            self.add_error('DB config id cannot be empty.')
            self.add_required('db_id')
        elif not regex_good_name.match(self.db_id):
            self.add_error(
                'Database ID may only contain letters, numbers and symbols specified in brackets [.-_@].'
            )

    def _validate_db_encoding(self):
        if not isinstance(self.db_encoding, basestring) or not self.db_encoding.strip():
            self.add_error('Database string encoding cannot be empty.')
            self.add_required('db_encoding')

    def _validate_autoload(self):
        if isinstance(self.autoload, bool):
            # keep boolean value
            pass
        elif str(self.autoload).strip().lower() == 'true':
            self.autoload = True
        elif str(self.autoload).strip().lower() == 'false':
            self.autoload = False
        else:
            self.add_error('Option automatically reflect DB expects True or False, got {}.'.format(self.autoload))

    def _validate_autoconnect(self):
        if isinstance(self.autoconnect, bool):
            # keep boolean value
            pass
        elif str(self.autoconnect).strip().lower() == 'true':
            self.autoconnect = True
        elif str(self.autoconnect).strip().lower() == 'false':
            self.autoconnect = False
        else:
            self.add_error(
                'Option connect automatically on start expects True of False, got {}.'.format(self.autoconnect))

    @abc.abstractmethod
    def validate(self, overridden_connection_string=False):
        """Validates config data."""
        return super(IDBConfig, self).validate()


def parse_as_list(permissions):
    """Converts given permissions to a list while removing duplicates.

    :param permissions: Iterable or string containing the permissions
    :returns tuple(list, bool)

    """
    parts = []
    is_invalid_type = False

    if not permissions:
        return parts, is_invalid_type

    if isinstance(permissions, basestring):
        parts = permissions.split(',')
    elif isinstance(permissions, collections.Iterable):
        parts = (i for i in permissions if isinstance(i, basestring))
    else:
        is_invalid_type = True

    def clean_parts():
        for part in parts:
            part = part.strip()
            if part != '':
                yield part

    # an ad hoc 'ordered set' based on OrderedDict().viewkeys()
    parts = collections.OrderedDict(
        (k, None) for k in clean_parts()
    ).keys()
    return parts, is_invalid_type
